use std::collections::HashMap;
use std::ffi::OsStr;
use std::path::{Component, Path};
use std::sync::atomic::{AtomicU64, Ordering};

use anyhow::Result;
use fuse::{
  FileAttr, FileType, Filesystem, ReplyAttr, ReplyData, ReplyDirectory,
  ReplyEmpty, ReplyEntry, ReplyOpen, Request,
};
use indexmap::IndexMap;
use time::Timespec;

type Ino = u64;

#[derive(Debug)]
pub enum FakeFile {
  Dir {
    ino: Ino,
    children: IndexMap<String, Ino>,
  },
  File {
    ino: Ino,
    size: u64,
  },
}

impl FakeFile {
  fn ino(&self) -> Ino {
    match self {
      FakeFile::Dir { ino, .. } | FakeFile::File { ino, .. } => *ino,
    }
  }

  fn size(&self) -> u64 {
    match self {
      FakeFile::Dir { .. } => 0,
      FakeFile::File { size, .. } => *size,
    }
  }

  fn ty(&self) -> FileType {
    match self {
      FakeFile::Dir { .. } => FileType::Directory,
      FakeFile::File { .. } => FileType::RegularFile,
    }
  }

  fn attr(&self) -> FileAttr {
    let time = Timespec::new(0, 0);
    FileAttr {
      ino: self.ino(),
      size: self.size(),
      blocks: 0,
      atime: time,
      mtime: time,
      ctime: time,
      crtime: time,
      kind: self.ty(),
      perm: 0o777,
      nlink: 0,
      uid: 0,
      gid: 0,
      rdev: 0,
      flags: 0,
    }
  }
}

#[derive(Debug)]
pub struct FakeFs {
  root: Ino,
  ctr: AtomicU64,
  info: HashMap<Ino, FakeFile>,
  handles: HashMap<Ino, Ino>,
}

impl FakeFs {
  pub fn new() -> Self {
    let ctr = AtomicU64::new(0);
    let root = ctr.fetch_add(1, Ordering::SeqCst);
    let mut info = HashMap::new();
    info.insert(
      root,
      FakeFile::Dir {
        ino: root,
        children: IndexMap::new(),
      },
    );
    let handles = HashMap::new();
    FakeFs {
      root,
      ctr,
      info,
      handles,
    }
  }

  fn inc(&self) -> u64 {
    self.ctr.fetch_add(1, Ordering::SeqCst)
  }

  pub fn set_size(&mut self, ino: Ino, new_size: u64) {
    let info = self.info.get_mut(&ino).unwrap();
    match info {
      FakeFile::File { size, .. } => *size = new_size,
      _ => {}
    }
  }

  pub fn ensure_path(
    &mut self,
    path: impl AsRef<Path>,
    ty: FileType,
  ) -> Result<Ino> {
    let path = path.as_ref();
    let parent_path = path.parent().unwrap();
    let filename = path.file_name().unwrap().to_str().unwrap().to_owned();
    let mut curr_ino = self.root;

    for component in parent_path.components() {
      let curr = self.info.get_mut(&curr_ino).unwrap();

      let name = match component {
        Component::RootDir => continue,
        Component::Normal(name) => name.to_str().unwrap().to_owned(),
        _ => continue,
      };

      let children = match curr {
        FakeFile::Dir { children, .. } => children,
        FakeFile::File { .. } => bail!("Trying to add children to file."),
      };

      if let Some(ino) = children.get(&name) {
        curr_ino = *ino;
      } else {
        let ino = self.ctr.fetch_add(1, Ordering::SeqCst);
        children.insert(name.clone(), ino);

        self.info.insert(
          ino,
          FakeFile::Dir {
            children: IndexMap::new(),
            ino,
          },
        );
      }
    }

    let curr = self.info.get_mut(&curr_ino).unwrap();
    let children = match curr {
      FakeFile::Dir { children, .. } => children,
      FakeFile::File { .. } => bail!("Trying to add children to file."),
    };

    let ino = self.ctr.fetch_add(1, Ordering::SeqCst);
    children.insert(filename.clone(), ino);
    let new = match ty {
      FileType::Directory => FakeFile::Dir {
        children: IndexMap::new(),
        ino,
      },
      FileType::RegularFile => FakeFile::File { size: 0, ino },
      _ => unreachable!(),
    };
    self.info.insert(ino, new);

    Ok(ino)
  }
}

impl Filesystem for FakeFs {
  fn access(
    &mut self,
    _req: &Request,
    _ino: u64,
    _mask: u32,
    reply: ReplyEmpty,
  ) {
    reply.ok();
  }

  fn getattr(&mut self, req: &Request, ino: Ino, reply: ReplyAttr) {
    info!("Requested attributes: {req:?}");
    info!("for inode {ino}");

    let file_info = match self.info.get(&ino) {
      Some(v) => v,
      None => {
        error!("No file at ino {ino}");
        return reply.error(2);
      }
    };
    info!("File info: {file_info:?}");

    let time = Timespec::new(0, 0);
    let attr = file_info.attr();
    reply.attr(&time, &attr);
  }

  fn read(
    &mut self,
    _req: &Request,
    _ino: u64,
    _fh: u64,
    _offset: i64,
    _size: u32,
    reply: ReplyData,
  ) {
    info!("ASFIJPWEOIMV");
  }

  fn lookup(
    &mut self,
    _req: &Request,
    parent: u64,
    name: &OsStr,
    reply: ReplyEntry,
  ) {
    let name = name.to_str().unwrap().to_owned();
    info!("Asked for opendir for parent {parent} and name {name}.");

    let file_info = match self.info.get(&parent) {
      Some(v) => v,
      None => return reply.error(2),
    };

    let children = match file_info {
      FakeFile::Dir { children, .. } => children,
      FakeFile::File { .. } => return reply.error(20),
    };

    match children.get(&name) {
      Some(v) => {
        let info = self.info.get(&v).unwrap();
        let time = Timespec::new(0, 0);
        reply.entry(&time, &info.attr(), self.inc())
      }
      None => return reply.error(2),
    }
  }

  fn opendir(
    &mut self,
    _req: &Request,
    ino: u64,
    flags: u32,
    reply: ReplyOpen,
  ) {
    info!("Asked for opendir for inode {ino}.");
    let fh = self.inc();
    self.handles.insert(fh, ino);
    reply.opened(fh, flags);
  }

  fn readdir(
    &mut self,
    _req: &Request,
    ino: u64,
    fh: u64,
    offset: i64,
    mut reply: ReplyDirectory,
  ) {
    let offset = offset as usize;
    info!("Asked for readdir for inode {ino} @ offset {offset}.");

    let file_info = match self.info.get(&ino) {
      Some(v) => v,
      None => return reply.error(2),
    };
    info!("File info: {file_info:?}");

    let children = match file_info {
      FakeFile::Dir { children, .. } => children,
      FakeFile::File { .. } => return reply.error(20),
    };

    if offset < children.len() {
      for (i, (child_name, child_ino)) in
        children.iter().skip(offset).enumerate()
      {
        let child = self.info.get(&child_ino).unwrap();
        reply.add(ino, (i + 1) as i64, child.ty(), child_name);
        println!("Returned {child_name}");
      }
    }

    reply.ok();
  }
}

#[derive(Debug)]
pub struct FsNode {
  ino: Ino,
}
